/* Copyright  1999 Jim Winstead Jr. All Rights Reserved. */
/* $Id: $ */

/* History:
	1.4
	+ TODO: 

    1.3 (Released February 1, 1999)
	+ Underline link in About box.
	+ Fixed problem with starting screensaver with double-click.
	+ Deal with Explorer restarting.

	1.2 (Released January 17, 1999)
	+ Fixed CPU usage.
	+ Added 'Configure Screensaver'

	1.1 (Released January 12, 1999)
	+ Made popup menu dynamic.
	+ Added Enable/Disable Screensaver menu item.

	1.0 (Released January 11, 1999)
	+ Initial version.

 */

#define TITLE	"Tarsier"
#define URL		"http://trainedmonkey.com/tarsier/"

#define DESCRIPTION \
	"The tarsier is the tiniest species of monkey.\015\012\015\012" \
	"This tarsier allows you to easily enable or disable " \
	"your screen saver and activate it, all from a single " \
	"icon in the system tray.\015\012\015\012" \
	"To enable or disable your screen saver, click the icon " \
	"(which will be a blue plus sign if the screen saver is " \
	"currently enabled, otherwise a red minus sign).\015\012\015\012" \
	"To start your screen saver (regardless of whether it is " \
	"enabled), double-click the icon.\015\012\015\012" \
	"For more information and updates, visit:"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shellapi.h>
#include "resource.h"

#define WM_TARSIER	(WM_APP)

HINSTANCE	t_instance	= (HINSTANCE)0;
HWND		t_window	= (HWND)0;
HWND		t_about		= (HWND)0;
HICON		activeicon	= (HICON)0;
HICON		inactiveicon= (HICON)0;
HFONT		t_linkfont	= (HFONT)0;
BOOL		sc_active	= false;

static bool CreateTrayIcon(HWND window);
static void UpdateTrayIcon(HWND window);
static void StartScreenSaver(void);
static void ToggleScreenSaver(HWND window);

/* The preview code doesn't work because GetMenuItemRect() is broken. */
#undef ENABLE_PREVIEW

#ifdef ENABLE_PREVIEW
HWND		t_preview	= (HWND)0;
HMENU		t_previewmenu = (HMENU)0;
HANDLE		t_previewprocess = (HANDLE)0;
static HWND PreviewScreenSaver(HINSTANCE instance, RECT rect);
#endif

void DisplayPopupMenu(HWND window) {
	HMENU popup = CreatePopupMenu();

	{ // ADD START MENU ITEM (long form so we can make it bold!)
		MENUITEMINFO item; ZeroMemory(&item, sizeof(item)); item.cbSize = sizeof(item);
		item.fMask = MIIM_ID|MIIM_STATE|MIIM_TYPE;
		item.fType = MFT_STRING;
		item.fState = MFS_DEFAULT;
		item.dwTypeData = "&Start Screensaver";
		item.cch = strlen(item.dwTypeData);
		item.wID = IDM_SAVE;
		InsertMenuItem(popup, 0, 1, &item);
	}

	if (sc_active) {
		AppendMenu(popup, MF_STRING, IDM_TOGGLE, "&Disable Screensaver");
	}
	else {
		AppendMenu(popup, MF_STRING, IDM_TOGGLE, "&Enable Screensaver");
	}

	AppendMenu(popup, MF_SEPARATOR, 0, NULL);
	AppendMenu(popup, MF_STRING, IDM_CONFIG, "&Configure Screensaver");
//	AppendMenu(popup, MF_STRING, IDM_SETTIME,"Set Screensaver &Timeout");

#ifdef ENABLE_PREVIEW
	t_previewmenu = CreatePopupMenu();
	AppendMenu(t_previewmenu, MF_OWNERDRAW, IDM_PREVIEW, NULL);
	AppendMenu(popup, MF_POPUP, (UINT)t_previewmenu, "&Preview Screensaver");
#endif

	AppendMenu(popup, MF_SEPARATOR, 0, NULL);
	AppendMenu(popup, MF_STRING, IDM_ABOUT, "&About Tarsier...");
	AppendMenu(popup, MF_STRING, IDM_EXIT, "E&xit");

	POINT point; GetCursorPos(&point);
	// This is required when using a notify icon -- see KB article 
	// PRB: Menus for Notification Icons Don't Work Correctly 
	SetForegroundWindow(window);
	if (!TrackPopupMenu(popup, TPM_RIGHTBUTTON, point.x, point.y, 0, window, NULL)) {
		MessageBox(window, "Unable to display popup menu.", TITLE, 0);
		DestroyWindow(window);
		return;
	}
	// This is required when using a notify icon -- see KB article 
	// PRB: Menus for Notification Icons Don't Work Correctly 
	PostMessage(window, WM_USER, 0, 0);
	DestroyMenu(popup);
#ifdef ENABLE_PREVIEW
	DestroyMenu(t_previewmenu); t_previewmenu = 0;
#endif
}

void ConfigScreenSaver(HWND window) {
	char saver[MAX_PATH];
	if (GetPrivateProfileString("boot", "SCRNSAVE.EXE", NULL, saver, sizeof(saver), "SYSTEM.INI")) {
		STARTUPINFO si; ZeroMemory(&si, sizeof(si)); si.cb = sizeof(si);
		PROCESS_INFORMATION pi; ZeroMemory(&pi, sizeof(pi));
		if (!CreateProcess(NULL, saver, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi)) {
			MessageBox(window, "Unable to start screensaver configuration.", TITLE, 0);
		}
	}
	else {
		MessageBox(window, "Unable to find screensaver.", TITLE, 0);
	}
}

BOOL CALLBACK AboutDialogProc(HWND dialog, UINT message, WPARAM wparam, LPARAM lparam) {
	switch (message) {
	case WM_INITDIALOG:
		{ // SET DESCRIPTION TEXT
			SetWindowText(GetDlgItem(dialog, IDC_DESCRIPTION), DESCRIPTION);
		}
		return 1;

	case WM_CTLCOLORSTATIC:
		{ // SET FONT FOR URL
			HWND control = (HWND)lparam;
			if (GetDlgCtrlID(control) == IDC_URL) {
				HDC dc = (HDC)wparam;
				if (!t_linkfont) {
					HFONT font = (HFONT)GetCurrentObject(dc, OBJ_FONT);
					if (font) {
						LOGFONT fontinfo;
						if (GetObject(font, sizeof(fontinfo), &fontinfo)) {
							fontinfo.lfUnderline = TRUE;
							t_linkfont = CreateFontIndirect(&fontinfo);
						}
					}
				}

				// SELECT UNDERLINE FONT
				if (t_linkfont) {
					SelectObject(dc, t_linkfont);
				}

				// CHANGE COLOR
				SetBkColor(dc, GetSysColor(COLOR_3DFACE));
				SetTextColor(dc, RGB(0,0,255));

				return (BOOL)GetStockObject(NULL_BRUSH);
			}
		}
		return 0;
	case WM_COMMAND:
		switch (LOWORD(wparam)) {
		case IDCANCEL:
			EndDialog(dialog, 0);
			break;
		case IDC_URL:
			{ // OPEN THE URL
				ShellExecute(NULL, "open", URL, NULL, NULL, SW_SHOW);
			}
			break;
		}
		return 0;
	}
	return 0;
}

LRESULT CALLBACK TarsierWindowProc(HWND window, UINT message, WPARAM wparam, LPARAM lparam) {
	static UINT taskbarrestart = 0;

	switch (message) {
	case WM_CREATE:
		taskbarrestart = RegisterWindowMessage(TEXT("TaskbarCreated"));
		break;

	case WM_ACTIVATE:
		if (LOWORD(wparam) != WA_INACTIVE) {
			SetForegroundWindow(window);
		}
		return 0;
	case WM_TARSIER:
		switch (lparam) {
		case WM_LBUTTONDBLCLK:
			ToggleScreenSaver(window); // UNDO WHAT WE DID ON FIRST CLICK
			StartScreenSaver();
			return 0;
		case WM_LBUTTONDOWN:
			ToggleScreenSaver(window);
			return 0;
		case WM_RBUTTONUP:
			DisplayPopupMenu(window);
			return 0;
		}
		return 1;

	case WM_COMMAND:
		switch (LOWORD(wparam)) {
		case IDM_TOGGLE:
			ToggleScreenSaver(window);
			return 0;
		case IDM_ABOUT:
			if (!t_about) {
				t_about = CreateDialog(t_instance, MAKEINTRESOURCE(IDD_ABOUT), window, AboutDialogProc);
			}
			ShowWindow(t_about, SW_SHOW);
			return 0;
		case IDM_SAVE:
			StartScreenSaver();
			return 0;
		case IDM_CONFIG:
			ConfigScreenSaver(window);
			return 0;
		case IDM_SETTIME:
			return 0;
		case IDM_EXIT:
			DestroyWindow(window);
			return 0;
		}
		return 1;

#ifdef ENABLE_PREVIEW
	case WM_MEASUREITEM:
		{
			LPMEASUREITEMSTRUCT mis = (LPMEASUREITEMSTRUCT)lparam;
			mis->itemWidth = 120;
			mis->itemHeight = 90;
		}
		return TRUE;

	case WM_DRAWITEM:
		{
//			LPDRAWITEMSTRUCT dis = (LPDRAWITEMSTRUCT)lparam;
			if (!t_preview) {
				RECT rect;
				if (GetMenuItemRect(window, t_previewmenu, 0, &rect)) {
					t_preview = PreviewScreenSaver(t_instance, rect);
				}
			}
		}
		return TRUE;

	case WM_MENUSELECT:
		if (t_preview) {
			HMENU menu = (HMENU)lparam;
			if (menu != t_previewmenu && LOWORD(wparam) != 5) {
				DestroyWindow(t_preview);
				t_preview = 0;
			}
			return TRUE;
		}
		break;
#endif

	case WM_DESTROY:
		PostQuitMessage(0);
		return 1;

	default:
		if (message == taskbarrestart) {
			CreateTrayIcon(window);
		}
		break;
	}
	
	return DefWindowProc(window,message,wparam,lparam);
}

bool RegisterWindowClasses(HINSTANCE instance) {
	{ // REGISTER THE MAIN WINDOW CLASS
		WNDCLASSEX wndclass;
		ZeroMemory(&wndclass, sizeof(wndclass));
		wndclass.cbSize = sizeof(wndclass);
		wndclass.style = CS_DBLCLKS;
		wndclass.lpfnWndProc = TarsierWindowProc;
		wndclass.hInstance = instance;
		wndclass.lpszClassName = TITLE"_CLASS";
		wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
		wndclass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
		if (!RegisterClassEx(&wndclass)) {
			return false;
		}
	}
		
#ifdef ENABLE_PREVIEW
	{ // REGISTER THE PREVIEW WINDOW CLASS
		WNDCLASSEX wndclass;
		ZeroMemory(&wndclass, sizeof(wndclass));
		wndclass.cbSize = sizeof(wndclass);
		wndclass.lpfnWndProc = PreviewWindowProc;
		wndclass.hInstance = instance;
		wndclass.lpszClassName = TITLE"_PREVIEW_CLASS";
		wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
		wndclass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
		if (!RegisterClassEx(&wndclass)) {
			return false;
		}
	}
#endif

	return true;
}

int WINAPI WinMain(HINSTANCE instance, HINSTANCE /*previous*/, LPSTR /*args*/, int /*show*/) {
	{ // SAVE THE INSTANCE
		t_instance		= instance;
	}

	if (!RegisterWindowClasses(instance)) {
		return FALSE;
	}

	{ // LOAD THE ICONS
		activeicon		= LoadIcon(instance, MAKEINTRESOURCE(IDI_ACTIVE));
		inactiveicon	= LoadIcon(instance, MAKEINTRESOURCE(IDI_INACTIVE));
	}
	
	{ // CREATE THE MAIN WINDOW
		t_window = CreateWindowEx(
						0,
						TITLE"_CLASS",
						TITLE,
						0, //WS_VISIBLE,
						0, 0,
						0, 0,
						NULL,
						NULL,
						instance,
						NULL);
		if (!t_window) {
			return FALSE;
		}
	}

	// GET SCREENSAVER STATUS
	SystemParametersInfo(SPI_GETSCREENSAVEACTIVE, 0, &sc_active, 0);

	if (!CreateTrayIcon(t_window)) {
		MessageBox(NULL, "Unable to create tray icon.", TITLE, 0);
		return -1;
	}

	{ // MESSAGE LOOP
		MSG message;
		for (;;) {
			if (!GetMessage(&message, (HWND)NULL, 0, 0)) {
				break;
			}
			TranslateMessage(&message);
			DispatchMessage(&message);
		}
	}

#ifdef ENABLE_PREVIEW
	{ // KILL THE PREVIEW PROCESS
		if (t_previewprocess) {
			TerminateProcess(t_previewprocess, 0);
		}
	}
#endif

	{ // DELETE THE SYSTEM TRAY ICON
		NOTIFYICONDATA nid;
		ZeroMemory(&nid, sizeof(nid)); nid.cbSize = sizeof(nid);
		nid.hWnd	= t_window;
		nid.uID		= 0;

		if (!Shell_NotifyIcon(NIM_DELETE, &nid)) {
			MessageBox(NULL, "Unable to delete tray icon.", TITLE, 0);
			return 1;
		}
	}

	{ // UNLOAD RESOURCES
		if (activeicon)		DestroyIcon(activeicon);
		if (inactiveicon)	DestroyIcon(inactiveicon);
		if (t_linkfont)		DeleteObject(t_linkfont);
	}

	return 0;
}

static bool CreateTrayIcon(HWND window) {
	NOTIFYICONDATA nid;
	ZeroMemory(&nid, sizeof(nid)); nid.cbSize = sizeof(nid);
	nid.hWnd	= window;
	nid.uID		= 0;
	nid.uFlags	= NIF_ICON|NIF_TIP|NIF_MESSAGE;
	nid.hIcon	= sc_active ? activeicon : inactiveicon;
	nid.uCallbackMessage = WM_TARSIER;
	if (sc_active) {
		strncpy(nid.szTip, "Disable Screensaver", sizeof(nid.szTip));
	}
	else {
		strncpy(nid.szTip, "Enable Screensaver", sizeof(nid.szTip));
	}

	if (!Shell_NotifyIcon(NIM_ADD, &nid)) {
		return false;
	}
	return true;
}

static void UpdateTrayIcon(HWND window) {
	// CHANGE ICON
	NOTIFYICONDATA nid;
	ZeroMemory(&nid, sizeof(nid)); nid.cbSize = sizeof(nid);
	nid.hWnd	= window;
	nid.uID		= 0;
	nid.uFlags	= NIF_ICON|NIF_TIP;
	nid.hIcon	= sc_active ? activeicon : inactiveicon;
	if (sc_active) {
		strncpy(nid.szTip, "Disable Screensaver", sizeof(nid.szTip));
	}
	else {
		strncpy(nid.szTip, "Enable Screensaver", sizeof(nid.szTip));
	}

	if (!Shell_NotifyIcon(NIM_MODIFY, &nid)) {
#if _DEBUG
		MessageBox(NULL, "Unable to update tray icon.", TITLE, 0);
#endif
	}
}

static void StartScreenSaver(void) {
	// TEMPORARILY ENABLE SCREENSAVER
	SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, TRUE, NULL, SPIF_UPDATEINIFILE|SPIF_SENDCHANGE);
	// START SCREENSAVER
	PostMessage(GetDesktopWindow(), WM_SYSCOMMAND, SC_SCREENSAVE, 0);
	// RESTORE SCREENSAVER STATUS
	if (!sc_active) { // only if setting inactive, otherwise things go wrong.
		SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, sc_active, NULL, SPIF_UPDATEINIFILE|SPIF_SENDCHANGE);
	}
}

void ToggleScreenSaver(HWND window) {
	// GET SCREENSAVER STATUS
	SystemParametersInfo(SPI_GETSCREENSAVEACTIVE, 0, &sc_active, 0);
	sc_active = !sc_active;

	// TOGGLE SCREENSAVER
	SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, sc_active, NULL, SPIF_UPDATEINIFILE|SPIF_SENDCHANGE);

	// UPDATE THE ICON
	UpdateTrayIcon(window);
}

#ifdef ENABLE_PREVIEW
static LRESULT CALLBACK PreviewWindowProc(HWND window, UINT message, WPARAM wparam, LPARAM lparam) {
	switch(message) {
	case WM_DESTROY:
		if (t_previewprocess) {
			TerminateProcess(t_previewprocess, 0);
			t_previewprocess = 0;
		}
		return 1;
	}
	return DefWindowProc(window,message,wparam,lparam);
}

#include <stdio.h>

static HWND PreviewScreenSaver(HINSTANCE instance, RECT rect) {
	char saver[MAX_PATH];
	if (GetPrivateProfileString("boot", "SCRNSAVE.EXE", NULL, saver, sizeof(saver), "SYSTEM.INI")) {
		HWND window = CreateWindowEx(
						WS_EX_TOPMOST,
						TITLE"_PREVIEW_CLASS",
						TITLE,
						WS_POPUP,
						rect.left, rect.top,
						rect.right-rect.left, rect.bottom-rect.top,
						NULL,
						NULL,
						instance,
						NULL);
		if (!window) {
			return 0;
		}
		ShowWindow(window, SW_SHOWNA);
		char command[MAX_PATH];
		STARTUPINFO si; ZeroMemory(&si, sizeof(si)); si.cb = sizeof(si);
		PROCESS_INFORMATION pi; ZeroMemory(&pi, sizeof(pi));
		_snprintf(command, sizeof(command), "%s /p %d", saver, window);
		if (!CreateProcess(NULL, command, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi)) {
			DestroyWindow(window);
		}
		t_previewprocess = pi.hProcess;
		return window;
	}
	else {
		MessageBox(NULL, "Unable to find screensaver.", TITLE, 0);
	}
	return 0;
}
#endif
